#!/usr/bin/env python3
import json
import re
import sys

def freeze(data):
    if isinstance(data, dict):
        return frozenset((k, freeze(v)) for k,v in data.items())
    elif isinstance(data, list):
        return frozenset(map(freeze, data))
    else:
        return data

with open(sys.argv[1], 'rb') as f:
    d = f.read()

with open(sys.argv[2], 'r') as f:
    expected = freeze(json.load(f))

if d[:7] == b'\x7fELF\x02\x01\x01':
    mode = 64
elif d[:7] == b'\x7fELF\x01\x01\x01':
    mode = 32
else:
    raise ValueError('not an elf64le nor elf32le')

if mode == 32:
    phoff = int.from_bytes(d[0x1c:0x20], 'little')
    phentsize = int.from_bytes(d[0x2a:0x2c], 'little')
    phnum = int.from_bytes(d[0x2c:0x2e], 'little')
else:
    phoff = int.from_bytes(d[0x20:0x28], 'little')
    phentsize = int.from_bytes(d[0x36:0x38], 'little')
    phnum = int.from_bytes(d[0x38:0x3a], 'little')

for i in range(phnum):
    pos = phoff + i * phentsize
    p_type = int.from_bytes(d[pos:pos+4], 'little')
    if p_type == 4:
        if mode == 64:
            note_start = int.from_bytes(d[pos+8:pos+16], 'little')
            note_size = int.from_bytes(d[pos+32:pos+40], 'little')
        else:
            note_start = int.from_bytes(d[pos+4:pos+8], 'little')
            note_size = int.from_bytes(d[pos+16:pos+20], 'little')
        note = d[note_start:note_start+note_size]
        break
else:
    raise ValueError('no NOTE')

pos = 0
res = []
while pos < len(note):
    namesz = int.from_bytes(note[pos:pos+4], 'little')
    descsz = int.from_bytes(note[pos+4:pos+8], 'little')
    type = int.from_bytes(note[pos+8:pos+12], 'little')
    pos += 12
    name = note[pos:pos+namesz]
    assert name[-1] == 0
    name = name[:-1].decode()
    pos += namesz
    while pos % 4: pos += 1
    desc = note[pos:pos+descsz]
    pos += descsz
    while pos % 4: pos += 1
    if name == 'CORE' and type == 1:
        pid = int.from_bytes(desc[0x20:0x24], 'little')
    elif name == 'LINUX' and type == 0x4f4453:
        num_fds = int.from_bytes(desc[:4], 'little')
        dpos = 4
        proc_fds = {}
        for _ in range(num_fds):
            fd = int.from_bytes(desc[dpos:dpos+4], 'little')
            fd_flags = int.from_bytes(desc[dpos+4:dpos+8], 'little')
            fd_pos = int.from_bytes(desc[dpos+8:dpos+16], 'little')
            fd_mode = int.from_bytes(desc[dpos+16:dpos+20], 'little')
            fd_domain = int.from_bytes(desc[dpos+20:dpos+24], 'little')
            fd_type = int.from_bytes(desc[dpos+24:dpos+28], 'little')
            local_addr_len = int.from_bytes(desc[dpos+28:dpos+30], 'little')
            remote_addr_len = int.from_bytes(desc[dpos+30:dpos+32], 'little')
            dpos += 32
            local_addr = desc[dpos:dpos+local_addr_len]
            dpos += local_addr_len
            remote_addr = desc[dpos:dpos+remote_addr_len]
            dpos += remote_addr_len
            end = desc.index(b'\x00', dpos)
            fd_path = desc[dpos:end].decode()
            dpos = end + 1
            while dpos % 4: dpos += 1

            path_pattern = re.sub('^(socket|pipe):\[[0-9]+\]$', '\g<1>:[XXXX]', fd_path)
            path_pattern = re.sub('^/dev/tty[0-9]+$', '/dev/ttyXXX', path_pattern)

            proc_fds[str(fd)] = {
                    'flags': fd_flags,
                    'pos': fd_pos,
                    'mode': fd_mode,
                    'path_pattern': path_pattern,
            }

            if fd_domain != 0xffffffff:
                proc_fds[str(fd)].update({
                    'domain': fd_domain,
                    'type': fd_type,
                    'local': local_addr.hex(),
                    'remote': remote_addr.hex(),
                })
        res.append(proc_fds)
        if dpos != len(desc):
            raise ValueError('fds desynced')

res = freeze(res)
if pos != len(note):
    raise ValueError('notes desynced')

print('\033[32mOK\033[0m' if res == expected else '\033[31mFAIL\033[0m')

